

clc;
clear;
close;
addpath('C:\aaa._MY_FILES\aUSN\4th semester\casadi-windows-matlabR2016a-v3.5.5');
import casadi.*

%Define requirements
% hg_max = 5.6;

%Level constraints from April 15 to May 15
x_LRV1 = 55.75; %[m] from april 15 to april 30    
x_HRV1 = 60.35; %[m] from april 15 to april 30

x_LRV2 = 58.85; % [m] from may 1st to may 15
x_HRV2 = 59.85; % [m] from may 1st to may 15

Vi_dot = 250; % Inflow variations in the lake Toke [m3/sec]

Vt_dot = 36;

x_LRV_min = 55.75; %[m] 

% Time setting for simulations
initial_time = 0;
final_time = 30*86400; % Total simulation period of 13 days
sampling_time = 3600;% Sampling time of 1 hour
t = initial_time:sampling_time:final_time;
number_timesteps = int64((final_time-initial_time)/sampling_time) + 1 ;
hours = 1:1:number_timesteps;

%%Creating array with constraints for plotting
xLRV = zeros(1,number_timesteps);
xLRV(1:360) = x_LRV1;
xLRV(360:end) = x_LRV2;

xHRV = zeros(1,number_timesteps);
xHRV(1:360) = x_HRV1;
xHRV(360:end) = x_HRV2;

%%Symbolic representation for casadi
%%States representation
h = MX.sym('h',2);
states = [h] ;
n_states = length(states);

%Control action representation
hg = MX.sym('hg',2,1);
controls = [hg];
n_control = length(controls);


%Process model
hdot = State_models(states,controls);

% Non Linear mapping function
f = Function('f',{states,controls},{hdot}); 


% For storing states of the system
% H = DM(n_states,number_timesteps);

H = SX.sym('H',n_states,number_timesteps);

% Control signal over the simulation period
Hg = DM(n_control,number_timesteps);
 
%States assignment
initial_states = [58.75;58.70];
H(:,1)= initial_states;
El_t = zeros(number_timesteps-1,1);
tic;
 for k = 1:number_timesteps  -1
%          tic;
        if k<=50
            Hg(:,k) = 0;
             
        elseif k>50 && k<=150
            Hg(:,k) = 0;            
            
        elseif k>150 && k<=340
            Hg(:,k) = 0;  
            
        elseif k>300 && k<=400
            Hg(:,k) = 1.2;
            
            
        elseif k>400 && k<=500
             Hg(:,k) = 0.4;
             
        elseif k>500 && k<=700
            Hg(:,k) = 0.4;
        else
            Hg(:,k) = 0.4;
               
        end
        % Integrating the models with explicit runge kutta method
         k1 = f(initial_states,Hg(:,k));
         k2 = f(initial_states+k1.*sampling_time/2,Hg(:,k));
         k3 = f(initial_states+k2.*sampling_time/2,Hg(:,k));
         k4 = f(initial_states+k3.*sampling_time,Hg(:,k));
         H(:,k+1) = initial_states +sampling_time/6*(k1+2.*k2+2.*k3+k4);
         initial_states = H(:,k+1);
         
%          El_t(k,1)=toc;
 end
 
toc;

H = full(evalf(H)); 
Hg = full(evalf(Hg)); 

%%Plot results showing the level changes x_M and x_D
subplot(3,1,1);
plot5 = plot(hours,H,'LineWidth',1.05); 
hold on;
plot6 = plot(hours,xLRV,hours,xHRV,'LineWidth',1.05);
combined_plot = [plot5;plot6];
legend(combined_plot,'x_M','x_D','xLRV','xHRV');

% xlabel('Time(hours)');
ylabel('Levels[m]');
title('Open loop simulation for x_M and x_D change')
grid on;

subplot(3,1,2);
plot(hours,Hg,'LineWidth',1.05);
ylim([-0.95,5]);
legend('hg1','hg2');
% xlabel('Time(hours)');
ylabel('control signal, hg[m]');
title('Control singal change over the period')
grid on;

subplot(3,1,3)
plot(hours,ones(1,length(hours)).*Vt_dot,hours,ones(1,length(hours)).*Vi_dot);
ylim([0,300]);
xlabel('Time(hours)');
ylabel('volumetric flow, m^3/s');
legend('Vt dot','Vi dot');
